from datetime import datetime
from pathlib import Path
from unittest.mock import call, patch

from arkindex_worker.worker import Element

SAMPLES = Path(__file__).absolute().parent / "samples"


@patch("worker_file_import.worker.check_image")
def test_get_images(check_image_mock, mock_worker):
    check_image_mock.return_value = [Path("/something.jpg")]
    test_datafile = {
        "id": "file1",
        "name": "file1.jpg",
        "local_path": "some/path/file1.jpg",
        "content_type": "image/jpeg",
    }
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "element": {"id": "folderid"},
    }
    mock_worker.get_images(test_datafile)

    assert check_image_mock.call_count == 1
    assert check_image_mock.call_args == call(test_datafile)


@patch("worker_file_import.worker.check_image")
def test_get_images_error(check_image_mock, mock_worker):
    check_image_mock.side_effect = FileNotFoundError
    mock_worker.process = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "element": {"id": "folderid"},
    }
    datafile = {"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"}

    assert mock_worker.get_images(datafile) == []

    assert check_image_mock.call_count == 1
    assert check_image_mock.call_args == call(datafile)


@patch("worker_file_import.worker.datetime")
def test_create_top_level_folder(datetime_mock, mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "corpus": "corpusid",
        "element": None,
        "folder_type": "folder",
        "element_type": "page",
        "files": ["file1", "file2"],
    }
    datetime_mock.now.return_value = datetime(2020, 2, 2, 20, 2)
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "folder",
            "name": "Import images 2020-02-02 20:02",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        },
        response={
            "id": "folderid",
            "type": "folder",
            "name": "Import images 2020-02-02 20:02",
        },
    )

    mock_worker.create_top_level_folder(
        [
            {"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"},
            {"id": "file2", "name": "file2.jpg", "content_type": "image/jpeg"},
        ]
    )
    assert mock_worker.process_information["element"] == {
        "id": "folderid",
        "type": "folder",
        "name": "Import images 2020-02-02 20:02",
    }


def test_create_top_level_folder_datafile_name(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "corpus": "corpusid",
        "element": None,
        "folder_type": "folder",
        "element_type": "page",
        "files": ["file1"],
    }
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "folder",
            "name": "file.jpg",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        },
        response={"id": "folderid", "type": "folder", "name": "file.jpg"},
    )

    mock_worker.create_top_level_folder(
        [{"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"}]
    )

    assert mock_worker.process_information["element"] == {
        "id": "folderid",
        "type": "folder",
        "name": "file.jpg",
    }


def test_create_top_level_folder_ignore_element(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "corpus": "corpusid",
        "element": {"id": "folderid", "name": "My folder"},
        "folder_type": "folder",
        "element_type": "page",
        "files": ["file1"],
    }
    mock_worker.create_top_level_folder([{"id": "file1", "name": "file.jpg"}])


def test_save_elements(responses, mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "element": {"id": "folderid", "type": "folder", "corpus": {"id": "corpusid"}},
        "folder_type": "folder",
        "element_type": "page",
    }

    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "29a26cadf018c8bd5eb4e2eca4d58683"},
        response={
            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "https://nerv.co.jp/puturl",
            "s3_url": "http://s3/200x200.jpg",
        },
    )
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "page",
            "name": "200x200.jpg",
            "polygon": None,
            "confidence": None,
            "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "parent": "folderid",
        },
        response={"id": "pageid", "type": "page", "name": "200x200.jpg"},
    )

    datafile = {
        "id": "file1",
        "name": "200x200.jpg",
        "s3_url": "http://s3/200x200.jpg",
    }
    pages = mock_worker.save_elements(
        datafile,
        [SAMPLES / "200x200.jpg"],
        parent_element=Element(
            {"id": "folderid", "type": "folder", "corpus": {"id": "corpusid"}}
        ),
    )

    assert pages == [{"id": "pageid", "type": "page", "name": "200x200.jpg"}]


def test_save_elements_top_level(responses, mock_worker):
    """
    Assert that save_elements creates folders even when a datafile has a single image
    if the process runs on a corpus, without a parent element
    """
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "element": None,
        "folder_type": "folder",
        "element_type": "page",
    }

    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "29a26cadf018c8bd5eb4e2eca4d58683"},
        response={
            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "https://nerv.co.jp/puturl",
            "s3_url": "http://s3/200x200.jpg",
        },
    )
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "folder",
            "name": "200x200.jpg",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "parent": None,
        },
        response={
            "id": "folderid",
            "type": "folder",
            "name": "200x200.jpg",
            "corpus": {"id": "corpusid"},
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "page",
            "name": "1",
            "polygon": None,
            "confidence": None,
            "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "parent": "folderid",
        },
        response={"id": "pageid", "type": "page", "name": "1"},
    )

    datafile = {
        "id": "file1",
        "name": "200x200.jpg",
        "s3_url": "http://s3/200x200.jpg",
    }
    elements = mock_worker.save_elements(
        datafile, [SAMPLES / "200x200.jpg"], parent_element=None
    )

    assert elements == [
        {
            "id": "folderid",
            "type": "folder",
            "name": "200x200.jpg",
            "corpus": {"id": "corpusid"},
        },
        {"id": "pageid", "type": "page", "name": "1"},
    ]
