import sys
from datetime import datetime
from logging import Logger, getLogger

from natsort import natsorted

from arkindex_worker.image import update_pillow_image_size_limit
from arkindex_worker.worker import BaseWorker, Element, ElementMixin
from worker_file_import.image import check_image, create_image
from worker_file_import.utils import delete_files, download_files, dump_json

logger: Logger = getLogger(__name__)


class ImportWorker(BaseWorker, ElementMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.elements = []

    def configure(self, *args, **kwargs):
        super().configure(*args, *kwargs)
        # We need more info than just what is serialized in a worker run's process
        process_details = self.api_client.request(
            "RetrieveProcess", id=self.process_information["id"]
        )
        self.process_information["files"] = process_details["files"]

    def create_top_level_folder(self, datafiles):
        """
        When a process runs without a parent element, files that have a single image
        (images and single-page PDFs) would get created without a parent folder, so they
        would be invisible to the user.
        Here, we create a top-level folder for any import with images only.
        `save_elements` includes will do it for PDFs, and `import_archive` for archives.
        This receives the datafiles as an argument to deduce the folder's name.
        """
        if self.process_information.get("element") or not all(
            datafile["content_type"].startswith("image/") for datafile in datafiles
        ):
            return

        if len(datafiles) == 1:
            parent_name = datafiles[0]["name"]
        else:
            parent_name = (
                f'Import images {datetime.now().isoformat(sep=" ", timespec="minutes")}'
            )

        self.process_information["element"] = self.api_client.request(
            "CreateElement",
            body={
                "corpus": self.process_information["corpus"],
                "type": self.process_information["folder_type"],
                "name": parent_name,
                "worker_run_id": self.worker_run_id,
            },
        )

    def get_images(self, datafile):
        """
        From a given DataFile, returns a list of paths to images
        compatible with an IIIF server and ML operations
        """
        try:
            if datafile["content_type"].startswith("image/"):
                return check_image(datafile)
        except Exception as e:
            logger.warning(f"Could not fetch images from {datafile['name']}: {e}")
            return []

        raise NotImplementedError

    def save_elements(self, datafile, image_paths, parent_element):
        # When there is a single image, create the element immediately without a parent
        # If the import runs without a parent folder (on a corpus directly), we will still create
        # a parent folder to avoid elements that are invisible in the API.
        if len(image_paths) == 1 and parent_element:
            return [
                self.create_sub_element(
                    parent_element,
                    type=self.process_information["element_type"],
                    name=datafile["name"],
                    image=create_image(self.api_client, image_paths[0]),
                    slim_output=False,
                )
            ]
        # Create a parent folder for a file with multiple images
        parent = self.api_client.request(
            "CreateElement",
            body={
                "parent": None if not parent_element else parent_element.id,
                "corpus": self.process_information["corpus"],
                "type": self.process_information["folder_type"],
                "name": datafile["name"],
                "worker_run_id": self.worker_run_id,
            },
        )
        parent_element = Element(parent)

        elements = [parent]
        assert parent is not None, "Cannot create an image as a top-level element."
        elements.extend(
            [
                self.create_sub_element(
                    parent_element,
                    type=self.process_information["element_type"],
                    name=str(i),
                    image=create_image(self.api_client, image_path),
                    slim_output=False,
                )
                for i, image_path in enumerate(image_paths, start=1)
            ]
        )
        return elements

    def import_datafile(self, datafile, parent_element) -> None:
        """
        Import a single DataFile into a parent folder, and add the created elements to `self.elements`.

        When a file has not been successfully imported in its entirety, exceptions may be raised.
        """
        logger.info(f"Fetching images for {datafile['name']}")
        images = self.get_images(datafile)
        if not images:
            return

        logger.info(f"Creating {len(images)} elements")
        new_elements = self.save_elements(datafile, images, parent_element)
        self.elements.extend(new_elements)

    @update_pillow_image_size_limit
    def run(self):
        self.configure()

        logger.info("Running import {!s}".format(self.process_information["id"]))

        logger.info("Downloading files...")
        datafiles = download_files(self.api_client, self.process_information["files"])
        if len(datafiles) > 0:
            logger.info(f"Downloaded {len(datafiles)} files")
        else:
            logger.error("No files downloaded")
            sys.exit(1)

        # Sort files by name as a human would expect: a1.jpg < a10.jpg
        datafiles = natsorted(datafiles, key=lambda datafile: datafile["name"])

        self.create_top_level_folder(datafiles)

        parent_element = None
        if self.process_information["element"]:
            parent_element = Element(self.process_information["element"])
            logger.info(f'Using folder {parent_element.id} "{parent_element.name}"')
            self.elements.append(
                {
                    "type": parent_element["type"],
                    "id": parent_element["id"],
                }
            )

        successful_datafiles = []
        for df in datafiles:
            try:
                self.import_datafile(df, parent_element)
            except Exception as e:
                logger.error(
                    f'Could not import file {df["name"]}: {e.__class__.__name__}: {str(e)}'
                )
                continue

            successful_datafiles.append(df)

        if len(self.elements) < 1:
            logger.error("No elements have been imported")
            sys.exit(1)

        logger.info(f"Imported {len(self.elements) - 1} elements")

        # The file import might create any amount of folders, but the first one
        # is always the topmost one. We look for it in the elements list then
        # update the Process's element so that the frontend can display
        # a "View element" button.open(SAMPLES / "600x600.jpg", "rb")
        created_folder = next(
            (
                element
                for element in self.elements
                if element["type"] == self.process_information["folder_type"]
            ),
            None,
        )
        if created_folder:
            self.api_client.request(
                "PartialUpdateProcess",
                id=self.process_information["id"],
                body={"element_id": created_folder["id"]},
            )

        logger.info("Deleting files…")
        delete_files(self.api_client, successful_datafiles)

        dump_json(self.elements, self.work_dir)


def main() -> None:
    ImportWorker(
        description="Worker to import files in various formats in Arkindex"
    ).run()


if __name__ == "__main__":
    main()
